# Generated by Django 2.2.19 on 2021-03-22 02:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eri", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="process",
            name="root_pipeline_id",
            field=models.CharField(db_index=True, max_length=33, verbose_name="根流程 ID"),
        ),
    ]



"""

Start the celery clock service from the Django management command.

"""
from __future__ import absolute_import, unicode_literals

from optparse import make_option as Option

from celery.bin import beat

from pipeline.management.commands.app import app
from pipeline.management.commands.base import CeleryCommand

beat = beat.beat(app=app)


class Command(CeleryCommand):
    """Run the celery periodic task scheduler."""

    help = 'Old alias to the "celery beat" command.'
    options = (
        Option("-A", "--app", default=None),
        Option("--broker", default=None),
        Option("--loader", default=None),
        Option("--config", default=None),
        Option("--workdir", default=None, dest="working_directory"),
        Option("--result-backend", default=None),
        Option("--no-color", "-C", action="store_true", default=None),
        Option("--quiet", "-q", action="store_true"),
    )
    if beat.get_options() is not None:
        options = options + CeleryCommand.options + beat.get_options()

    def handle(self, *args, **options):
        beat.run(*args, **options)



# Generated by Django 2.2.6 on 2020-12-16 02:56

from django.db import migrations


def reverse_func(apps, schema_editor):
    raise Exception("task migrate cannot reverse")


def forward_func(apps, schema_editor):
    from pipeline.contrib.periodic_task.djcelery import migrate

    migrate.try_to_migrate_to_django_celery_beat()


class Migration(migrations.Migration):

    dependencies = [
        ("periodic_task", "0004_auto_20191213_0828"),
    ]

    operations = [migrations.RunPython(forward_func, reverse_func)]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json
from typing import Dict, List, Set

from django.db import transaction

from bamboo_engine import metrics
from bamboo_engine.eri import ContextValue, ContextValueType

from pipeline.eri.models import ContextValue as DBContextValue
from pipeline.eri.models import ContextOutputs
from pipeline.eri.imp.serializer import SerializerMixin


class ContextMixin(SerializerMixin):
    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_CONTEXT_VALUE_READ_TIME)
    def get_context_values(self, pipeline_id: str, keys: set) -> List[ContextValue]:
        """
        获取某个流程上下文中的 keys 所指定的键对应变量的值

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param keys: 变量键
        :type keys: set
        :return: 变量值信息
        :rtype: List[ContextValue]
        """
        qs = DBContextValue.objects.filter(pipeline_id=pipeline_id, key__in=keys).only(
            "key", "type", "serializer", "value", "code"
        )

        return [
            ContextValue(
                key=cv_model.key,
                type=ContextValueType(cv_model.type),
                value=self._deserialize(cv_model.value, cv_model.serializer),
                code=cv_model.code or None,
            )
            for cv_model in qs
        ]

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_CONTEXT_REF_READ_TIME)
    def get_context_key_references(self, pipeline_id: str, keys: set) -> set:
        """
        获取某个流程上下文中 keys 所指定的变量直接和间接引用的其他所有变量的键

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param keys: 变量 key 列表
        :type keys: set
        :return: keys 所指定的变量直接和简介引用的其他所有变量的键
        :rtype: set
        """
        qs = DBContextValue.objects.filter(pipeline_id=pipeline_id, key__in=keys).only("references")

        references = []
        for cv_model in qs:
            references.extend(json.loads(cv_model.references))

        return set(references)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_CONTEXT_VALUE_UPSERT_TIME)
    @transaction.atomic
    def upsert_plain_context_values(self, pipeline_id: str, update: Dict[str, ContextValue]):
        """
        更新或创建新的普通上下文数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :param update: 更新数据
        :type update: Dict[str, ContextValue]
        """
        exist_keys = DBContextValue.objects.filter(pipeline_id=pipeline_id).values_list("key", flat=True)
        update_keys = set(update.keys()).intersection(exist_keys)

        # update
        for k in update_keys:
            context_value = update[k]
            value, serializer = self._serialize(context_value.value)

            DBContextValue.objects.filter(pipeline_id=pipeline_id, key=k).update(
                type=ContextValueType.PLAIN.value, value=value, serializer=serializer, code="", references="[]",
            )

        # insert
        insert_keys = set(update.keys()).difference(exist_keys)
        context_value_models = []
        for k in insert_keys:
            context_value = update[k]
            value, serializer = self._serialize(context_value.value)

            context_value_models.append(
                DBContextValue(
                    pipeline_id=pipeline_id,
                    key=context_value.key,
                    type=ContextValueType.PLAIN.value,
                    serializer=serializer,
                    value=value,
                    code="",
                    references="[]",
                )
            )

        DBContextValue.objects.bulk_create(context_value_models, batch_size=500)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_CONTEXT_VALUE_READ_TIME)
    def get_context(self, pipeline_id: str) -> List[ContextValue]:
        """
        获取某个流程的所有上下文数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: [description]
        :rtype: List[ContextValue]
        """
        qs = DBContextValue.objects.filter(pipeline_id=pipeline_id).only("key", "type", "serializer", "value", "code")

        return [
            ContextValue(
                key=cv_model.key,
                type=ContextValueType(cv_model.type),
                value=self._deserialize(cv_model.value, cv_model.serializer),
                code=cv_model.code or None,
            )
            for cv_model in qs
        ]

    def get_context_outputs(self, pipeline_id: str) -> Set[str]:
        """
        获取流程上下文需要输出的数据

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: 输出数据 key
        :rtype: Set[str]
        """
        co_model = ContextOutputs.objects.get(pipeline_id=pipeline_id)
        return set(json.loads(co_model.outputs))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import pkgutil
import os
import sys
from importlib import import_module


logger = logging.getLogger("root")


def find_all_modules(module_dir, sub_dir=None):
    modules = []
    for _, name, is_pkg in pkgutil.iter_modules([module_dir]):
        if name.startswith("_"):
            continue
        module = name if sub_dir is None else "{}.{}".format(sub_dir, name)
        if is_pkg:
            modules += find_all_modules(os.path.join(module_dir, name), module)
        else:
            modules.append(module)
    return modules


def autodiscover_items(module):
    """
    Given a path to discover, auto register all items
    """
    # Workaround for a Python 3.2 bug with pkgutil.iter_modules
    module_dir = module.__path__[0]
    sys.path_importer_cache.pop(module_dir, None)
    modules = find_all_modules(module_dir)
    for name in modules:
        module_path = "{}.{}".format(module.__name__, name)
        try:
            __import__(module_path)
        except Exception as e:
            logger.error(f"[!] module({module_path}) import failed with err: {e}")


def autodiscover_collections(path):
    """
    Auto-discover INSTALLED_APPS modules and fail silently when
    not present. This forces an import on them to register any admin bits they
    may want.
    """
    from django.apps import apps

    for app_config in apps.get_app_configs():
        # Attempt to import the app's module.
        try:

            _module = import_module("%s.%s" % (app_config.name, path))
            autodiscover_items(_module)
        except ImportError as e:
            if not str(e) == "No module named %s" % path:
                pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from typing import Dict, List, Optional

from django.utils import timezone
from pipeline.eri.models import State as DBState
from pipeline.eri.signals import post_set_state

from bamboo_engine import metrics, states
from bamboo_engine.eri import State
from bamboo_engine.exceptions import StateVersionNotMatchError
from bamboo_engine.utils.string import unique_id


class StateMixin:
    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_STATE_READ_TIME)
    def get_state(self, node_id: str) -> State:
        """
        获取某个节点的状态对象

        : param node_id: 节点 ID
        : type node_id: str
        : return: State 实例
        : rtype: State
        """
        state = DBState.objects.get(node_id=node_id)

        return State(
            node_id=state.node_id,
            root_id=state.root_id,
            parent_id=state.parent_id,
            name=state.name,
            version=state.version,
            loop=state.loop,
            inner_loop=state.inner_loop,
            retry=state.retry,
            skip=state.skip,
            error_ignored=state.error_ignored,
            created_time=state.created_time,
            started_time=state.started_time,
            archived_time=state.archived_time,
        )

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_STATE_READ_TIME)
    def get_state_or_none(self, node_id: str) -> Optional[State]:
        """
        获取某个节点的状态对象，如果不存在则返回 None

        : param node_id: 节点 ID
        : type node_id: str
        : return: State 实例
        : rtype: State
        """
        try:
            return self.get_state(node_id)
        except DBState.DoesNotExist:
            return None

    def get_state_by_root(self, root_id: str) -> List[State]:
        """
        根据根节点 ID 获取一批节点状态

        :param root_id: 根节点 ID
        :type root_id: str
        :return: 节点状态列表
        :rtype: List[State]
        """
        qs = DBState.objects.filter(root_id=root_id)

        return [
            State(
                node_id=state.node_id,
                root_id=state.root_id,
                parent_id=state.parent_id,
                name=state.name,
                version=state.version,
                loop=state.loop,
                inner_loop=state.inner_loop,
                retry=state.retry,
                skip=state.skip,
                error_ignored=state.error_ignored,
                created_time=state.created_time,
                started_time=state.started_time,
                archived_time=state.archived_time,
            )
            for state in qs
        ]

    def get_state_by_parent(self, parent_id: str) -> List[State]:
        """
        根据父节点 ID 获取一批节点状态

        :param parent_id: 父节点 ID
        :type parent_id: str
        :return: 节点状态列表
        :rtype: List[State]
        """
        qs = DBState.objects.filter(parent_id=parent_id)

        return [
            State(
                node_id=state.node_id,
                root_id=state.root_id,
                parent_id=state.parent_id,
                name=state.name,
                version=state.version,
                loop=state.loop,
                inner_loop=state.inner_loop,
                retry=state.retry,
                skip=state.skip,
                error_ignored=state.error_ignored,
                created_time=state.created_time,
                started_time=state.started_time,
                archived_time=state.archived_time,
            )
            for state in qs
        ]

    def batch_get_state_name(self, node_id_list: List[str]) -> Dict[str, str]:
        """
        批量获取一批节点的状态

        :param node_id_list: 节点 ID 列表
        :type node_id_list: List[str]
        :return: 节点ID -> 状态名称
        :rtype: Dict[str, str]
        """
        qs = DBState.objects.filter(node_id__in=node_id_list).only("node_id", "name")
        return {state.node_id: state.name for state in qs}

    def has_state(self, node_id: str) -> bool:
        """
        是否存在某个节点的的状态

        :param node_id: 节点 ID
        :type node_id: str
        :return: 该节点状态是否存在
        :rtype: bool
        """
        return DBState.objects.filter(node_id=node_id).exists()

    def reset_state_inner_loop(self, node_id: str) -> int:
        """
        设置节点的当前流程重入次数

        :param node_id: 节点 ID
        :type node_id: str
        :return: 更新状态行数
        :rtype: int
        """
        return DBState.objects.filter(node_id=node_id).update(inner_loop=0)

    def reset_children_state_inner_loop(self, node_id: str) -> int:
        """
        批量设置子流程节点的所有子节点inner_loop次数

        :param node_id: 子流程节点 ID
        :type node_id: str
        :return: 更新状态行数
        :rtype: int
        """
        return DBState.objects.filter(parent_id=node_id).update(inner_loop=0)

    def set_state_root_and_parent(self, node_id: str, root_id: str, parent_id: str):
        """
        设置节点的根流程和父流程 ID

        :param node_id: 节点 ID
        :type node_id: str
        :param root_id: 根流程 ID
        :type root_id: str
        :param parent_id: 父流程 ID
        :type parent_id: str
        """
        DBState.objects.filter(node_id=node_id).update(root_id=root_id, parent_id=parent_id)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_STATE_WRITE_TIME)
    def set_state(
        self,
        node_id: str,
        to_state: str,
        version: Optional[str] = None,
        loop: int = -1,
        inner_loop: int = -1,
        root_id: Optional[str] = None,
        parent_id: Optional[str] = None,
        is_retry: bool = False,
        is_skip: bool = False,
        reset_retry: bool = False,
        reset_skip: bool = False,
        error_ignored: bool = False,
        reset_error_ignored: bool = False,
        refresh_version: bool = False,
        clear_started_time: bool = False,
        set_started_time: bool = False,
        clear_archived_time: bool = False,
        set_archive_time: bool = False,
        ignore_boring_set: bool = False,
        send_post_set_state_signal: bool = True,
    ) -> str:
        """
        设置节点的状态，如果节点存在，进行状态转换时需要满足状态转换状态机

        :param node_id: 节点 ID
        :type node_id: str
        :param to_state: 目标状态
        :type to_state: str
        :param loop: 循环次数, 为 -1 时表示不设置
        :type loop: int, optional
        :param inner_loop: 当前流程循环次数, 为 -1 时表示不设置
        :type inner_loop: int, optional
        :param version: 目标状态版本，为空时表示不做版本校验
        :type version: Optional[str], optional
        :param root_id: 根节点 ID，为空时表示不设置
        :type root_id: Optional[str], optional
        :param parent_id: 父节点 ID，为空时表示不设置
        :type parent_id: Optional[str], optional
        :param is_retry: 是否增加重试次数
        :type is_retry: bool, optional
        :param is_skip: 是否将跳过设置为 True
        :type is_skip: bool, optional
        :param reset_retry: 是否重置重试次数
        :type reset_retry: bool, optional
        :param reset_skip: 是否重置跳过标志
        :type reset_skip: bool, optional
        :param error_ignored: 是否为忽略错误跳过
        :type error_ignored: bool, optional
        :param reset_error_ignored: 是否重置忽略错误标志
        :type reset_error_ignored: bool, optional
        :param refresh_version: 是否刷新版本号
        :type refresh_version: bool, optional
        :param clear_started_time: 是否清空开始时间
        :type clear_started_time: bool, optional
        :param set_started_time: 是否设置开始时间
        :type set_started_time: bool, optional
        :param clear_archived_time: 是否清空归档时间
        :type clear_archived_time: bool, optional
        :param set_archive_time: 是否设置归档时间
        :type set_archive_time: bool, optional
        :param ignore_boring_set: 当 version 与 to_state 与当前实际状态一致时，是否忽略本次设置
        :param send_post_set_state_signal: 节点状态改变时，是否发送post_set_state信号
        :type send_post_set_state_signal: bool, optional
        :return: 该节点最新版本
        :rtype: str
        """
        state = self.get_state_or_none(node_id)
        ret_version = ""

        if state and version and state.version != version:
            raise StateVersionNotMatchError("state version({}) not match {}".format(state.version, version))

        # 只有在当前状态和版本一致的情况下允许实现幂等
        if ignore_boring_set and state and version and state.version == version and state.name == to_state:
            return state.version

        fields = {}

        if loop != -1:
            fields["loop"] = loop

        if inner_loop != -1:
            fields["inner_loop"] = inner_loop

        if root_id:
            fields["root_id"] = root_id

        if parent_id:
            fields["parent_id"] = parent_id

        if is_retry and state:
            fields["retry"] = state.retry + 1

        if is_skip and state:
            fields["skip"] = True

        if reset_retry and state:
            fields["retry"] = 0

        if reset_skip and state:
            fields["skip"] = False

        if reset_error_ignored and state:
            fields["error_ignored"] = False

        if error_ignored and state:
            fields["error_ignored"] = True

        if refresh_version or state is None:
            fields["version"] = unique_id("v")

        if clear_started_time and state:
            fields["started_time"] = None

        if set_started_time:
            fields["started_time"] = timezone.now()

        if clear_archived_time and state:
            fields["archived_time"] = timezone.now()

        if set_archive_time:
            fields["archived_time"] = timezone.now()

        if state:
            if not states.can_transit(from_state=state.name, to_state=to_state):
                raise RuntimeError(
                    "can't not transit node({}) state from {} to {}".format(node_id, state.name, to_state)
                )

            filters = {"node_id": node_id}
            if version:
                filters["version"] = version

            rows = DBState.objects.filter(**filters).update(name=to_state, **fields)

            if rows != 1:
                raise StateVersionNotMatchError("state with version({}) not exist".format(version))
            ret_version = fields.get("version", state.version)
        else:
            state = DBState.objects.create(node_id=node_id, name=to_state, **fields)
            ret_version = fields["version"]
        if send_post_set_state_signal:
            post_set_state.send(
                sender=DBState,
                node_id=node_id,
                to_state=to_state,
                version=ret_version,
                root_id=state.root_id,
                parent_id=state.parent_id,
                loop=loop,
            )
        return ret_version



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json
from typing import List, Optional, Dict

from django.utils import timezone
from django.db.models import F
from django.db import transaction

from bamboo_engine import metrics
from bamboo_engine.eri import ProcessInfo, SuspendedProcessInfo, DispatchProcess

from pipeline.eri.models import Process


class ProcessMixin:
    def beat(self, process_id: int):
        """
        进程心跳

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(last_heartbeat=timezone.now())

    def wake_up(self, process_id: int):
        """
        将当前进程标记为唤醒状态

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(asleep=False)

    def sleep(self, process_id: int):
        """
        将当前进程标记为睡眠状态

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(asleep=True)

    def suspend(self, process_id: int, by: str):
        """
        将当前进程标记为阻塞状态

        :param process_id: 进程 ID
        :type process_id: int
        :param by: 造成阻塞的节点信息
        :type by: str
        """
        Process.objects.filter(id=process_id).update(suspended=True, suspended_by=by)

    def resume(self, process_id: int):
        """
        将进程标记为非阻塞状态

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(suspended=False, suspended_by="")

    def batch_resume(self, process_id_list: List[int]):
        """
        批量将进程标记为非阻塞状态

        :param process_id_list: 进程 ID 列表
        :type process_id_list: List[int]
        """
        Process.objects.filter(id__in=process_id_list).update(suspended=False, suspended_by="")

    def die(self, process_id: int):
        """
        将当前进程标记为非存活状态

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(dead=True)

    @metrics.setup_histogram(metrics.ENGINE_RUNTIME_PROCESS_READ_TIME)
    def get_process_info(self, process_id: int) -> ProcessInfo:
        """
        获取某个进程的基本信息

        :param process_id: 进程 ID
        :type process_id: int
        :return: 进程基本信息
        :rtype: ProcessInfo
        """
        qs = Process.objects.filter(id=process_id).only(
            "id", "destination_id", "root_pipeline_id", "pipeline_stack", "parent_id"
        )

        if len(qs) != 1:
            raise Process.DoesNotExist("Process with id({}) does not exist".format(process_id))

        process = qs[0]
        return ProcessInfo(
            process_id=process.id,
            destination_id=process.destination_id,
            root_pipeline_id=process.root_pipeline_id,
            pipeline_stack=json.loads(process.pipeline_stack),
            parent_id=process.parent_id,
        )

    def kill(self, process_id: int):
        """
        强制结束某个进程正在进行的活动，并将其标志为睡眠状态

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(asleep=True)

    def get_suspended_process_info(self, suspended_by: str) -> List[SuspendedProcessInfo]:
        """
        获取由于 pipeline 暂停而被暂停执行的进程信息

        : param suspended_by: 进程 ID
        : type suspended_by: str
        : return: 暂停的进程信息
        : rtype: SuspendedProcessInfo
        """
        qs = Process.objects.filter(suspended_by=suspended_by).only(
            "id", "current_node_id", "root_pipeline_id", "pipeline_stack"
        )

        return [
            SuspendedProcessInfo(
                process_id=p.id,
                current_node=p.current_node_id,
                root_pipeline_id=p.root_pipeline_id,
                pipeline_stack=json.loads(p.pipeline_stack),
            )
            for p in qs
        ]

    def get_sleep_process_info_with_current_node_id(self, node_id: str) -> Optional[ProcessInfo]:
        """
        获取由于处于睡眠状态且当前节点 ID 为 node_id 的进程 ID

        : param node_id: 节点 ID
        : type node_id: str
        : return: 进程 ID
        : rtype: str
        """
        qs = Process.objects.filter(asleep=True, current_node_id=node_id).only(
            "id", "destination_id", "root_pipeline_id", "pipeline_stack", "parent_id"
        )

        if len(qs) == 0:
            return None

        if len(qs) != 1:
            raise ValueError("found multiple sleep process({}) with current_node_id({})".format(qs, node_id))

        return ProcessInfo(
            process_id=qs[0].id,
            destination_id=qs[0].destination_id,
            root_pipeline_id=qs[0].root_pipeline_id,
            pipeline_stack=json.loads(qs[0].pipeline_stack),
            parent_id=qs[0].parent_id,
        )

    def get_process_id_with_current_node_id(self, node_id: str) -> Optional[str]:
        """
        获取当前节点 ID 为 node_id 且存活的进程 ID

        : param node_id: 节点 ID
        : type node_id: str
        : return: 进程 ID
        : rtype: str
        """
        qs = Process.objects.filter(dead=False, current_node_id=node_id).only("id")

        if len(qs) == 0:
            return None

        if len(qs) != 1:
            raise ValueError("found multiple process({}) with current_node_id({})".format(qs, node_id))

        return qs[0].id

    def set_current_node(self, process_id: int, node_id: str):
        """
        将进程当前处理节点标记为 node

        :param process_id: 进程 ID
        :type process_id: int
        :param node_id: 节点 ID
        :type node_id: str
        """
        Process.objects.filter(id=process_id).update(current_node_id=node_id)

    def child_process_finish(self, parent_id: int, process_id: int) -> bool:
        """
        标记某个进程的子进程执行完成，并返回是否能够唤醒父进程继续执行的标志位

        :param parent_id: 父进程 ID
        :type parent_id: int
        :param process_id: 子进程 ID
        :type process_id: int
        :return: 是否能够唤醒父进程继续执行
        :rtype: bool
        """
        with transaction.atomic():
            Process.objects.filter(id=process_id).update(dead=True)

            Process.objects.filter(id=parent_id).update(ack_num=F("ack_num") + 1)

            # compare(where) and set(update)
            waked = Process.objects.filter(id=parent_id, ack_num=F("need_ack")).update(ack_num=0, need_ack=-1)

        return waked != 0

    def is_frozen(self, process_id: int) -> bool:
        """
        检测当前进程是否需要被冻结

        :param process_id: 进程 ID
        :type process_id: int
        :return: 是否需要被冻结
        :rtype: bool
        """
        return Process.objects.filter(id=process_id, frozen=True).exists()

    def freeze(self, process_id: int):
        """
        冻结当前进程

        :param process_id: 进程 ID
        :type process_id: int
        """
        Process.objects.filter(id=process_id).update(frozen=True)

    def fork(
        self,
        parent_id: str,
        root_pipeline_id: str,
        pipeline_stack: List[str],
        from_to: Dict[str, str],
    ) -> List[DispatchProcess]:
        """
        根据当前进程 fork 出多个子进程

        :param parent_id: 父进程 ID
        :type parent_id: str
        :param root_pipeline_id: 根流程 ID
        :type root_pipeline_id: str
        :param pipeline_stack: 子流程栈
        :type pipeline_stack: List[str]
        :param from_to: 子进程的执行开始节点和目标节点
        :type from_to: Dict[str, str]
        :return: 待调度进程信息列表
        :rtype: List[DispatchProcess]
        """
        qs = Process.objects.filter(id=parent_id).only("priority", "queue")
        stack_json = json.dumps(pipeline_stack)

        if not qs:
            raise Process.DoesNotExist("Process with id({}) does not exist".format(parent_id))

        children = [
            Process(
                parent_id=parent_id,
                asleep=True,
                destination_id=destination,
                current_node_id=current_node,
                root_pipeline_id=root_pipeline_id,
                pipeline_stack=stack_json,
                priority=qs[0].priority,
                queue=qs[0].queue,
            )
            for current_node, destination in from_to.items()
        ]

        Process.objects.bulk_create(children, batch_size=500)

        qs = Process.objects.filter(parent_id=parent_id, dead=False).only("id", "current_node_id")

        children_count = len(qs)
        expect = len(from_to)
        if children_count != expect:
            raise ValueError(
                "process({}) fork failed, children count({}) does not match expect({})".format(
                    parent_id, children_count, expect
                )
            )

        return [DispatchProcess(process_id=p.id, node_id=p.current_node_id) for p in qs]

    def join(self, process_id: int, children_id: List[str]):
        """
        让父进程等待子进程

        :param process_id: 父进程 ID
        :type process_id: int
        :param children_id: 子进程 ID 列表
        :type children_id: List[str]
        """
        Process.objects.filter(id=process_id).update(ack_num=0, need_ack=len(children_id))

    def set_pipeline_stack(self, process_id: int, stack: List[str]):
        """
        设置进程的流程栈

        :param process_id: 进程 ID
        :type process_id: int
        :param stack: 流程栈
        :type stack: List[str]
        """
        Process.objects.filter(id=process_id).update(pipeline_stack=json.dumps(stack))

    def get_process_info_with_root_pipeline(self, pipeline_id: str) -> List[ProcessInfo]:
        """
        根据根流程 ID 获取一批进程的信息

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        :return: 进程基本信息
        :rtype: List[ProcessInfo]
        """
        qs = Process.objects.filter(root_pipeline_id=pipeline_id).only(
            "id", "destination_id", "root_pipeline_id", "pipeline_stack", "parent_id"
        )

        return [
            ProcessInfo(
                process_id=process.id,
                destination_id=process.destination_id,
                root_pipeline_id=process.root_pipeline_id,
                pipeline_stack=json.loads(process.pipeline_stack),
                parent_id=process.parent_id,
            )
            for process in qs
        ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import json

from typing import Any

from django.utils.module_loading import import_string

DATA_JSON_ENCODER_PATH = None
DATA_JSON_OBJECT_HOOK_PATH = None

_LOCAL = {}


def _get_local(key, path):
    if key in _LOCAL:
        return _LOCAL[key]

    try:
        _LOCAL[key] = import_string(path)
    except ImportError:
        _LOCAL[key] = None

    return _LOCAL[key]


def _get_data_json_encoder():
    if not DATA_JSON_ENCODER_PATH:
        return None
    return _get_local("data_json_encoder", DATA_JSON_ENCODER_PATH)


def _get_data_json_object_hook():
    if not DATA_JSON_OBJECT_HOOK_PATH:
        return None
    return _get_local("data_json_object_hook", DATA_JSON_OBJECT_HOOK_PATH)


def data_json_loads(data: str) -> Any:
    return json.loads(data, object_hook=_get_data_json_object_hook())


def data_json_dumps(data: Any) -> str:
    return json.dumps(data, cls=_get_data_json_encoder())



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import logging
from typing import Optional

from celery import task
from celery.decorators import periodic_task
from celery.schedules import crontab
from django.conf import settings

from bamboo_engine.eri import ExecuteInterruptPoint, ScheduleInterruptPoint
from bamboo_engine.engine import Engine
from bamboo_engine.interrupt import (
    ExecuteInterrupter,
    ExecuteKeyPoint,
    ScheduleInterrupter,
    ScheduleKeyPoint,
)

from pipeline.eri.models import LogEntry
from pipeline.eri.runtime import BambooDjangoRuntime


logger = logging.getLogger(__name__)


@task(ignore_result=True)
def execute(
    process_id: int,
    node_id: str,
    root_pipeline_id: str = None,
    parent_pipeline_id: str = None,
    recover_point: str = None,
):
    runtime = BambooDjangoRuntime()
    recover_point = ExecuteInterruptPoint.from_json(recover_point)
    interrupter = ExecuteInterrupter(
        runtime=runtime,
        current_node_id=node_id,
        process_id=process_id,
        parent_pipeline_id=parent_pipeline_id,
        root_pipeline_id=root_pipeline_id,
        check_point=ExecuteInterruptPoint(name=ExecuteKeyPoint.ENTRY),
        recover_point=recover_point,
    )

    Engine(runtime).execute(
        process_id=process_id,
        node_id=node_id,
        root_pipeline_id=root_pipeline_id,
        parent_pipeline_id=parent_pipeline_id,
        interrupter=interrupter,
    )


@task(ignore_result=True)
def schedule(
    process_id: int, node_id: str, schedule_id: str, callback_data_id: Optional[int], recover_point: str = None
):
    runtime = BambooDjangoRuntime()
    recover_point = ScheduleInterruptPoint.from_json(recover_point)

    interrupter = ScheduleInterrupter(
        runtime=runtime,
        process_id=process_id,
        current_node_id=node_id,
        schedule_id=schedule_id,
        callback_data_id=callback_data_id,
        check_point=ScheduleInterruptPoint(name=ScheduleKeyPoint.ENTRY),
        recover_point=recover_point,
    )
    Engine(runtime).schedule(
        process_id=process_id,
        node_id=node_id,
        schedule_id=schedule_id,
        interrupter=interrupter,
        callback_data_id=callback_data_id,
    )


@periodic_task(run_every=(crontab(minute=0, hour=0)), ignore_result=True)
def clean_expired_log():
    expired_interval = getattr(settings, "LOG_PERSISTENT_DAYS", None)

    if expired_interval is None:
        expired_interval = 30
        logger.warning("LOG_PERSISTENT_DAYS are not found in settings, use default value: 30")

    del_num = LogEntry.objects.delete_expired_log(expired_interval)
    logger.info("%s log entry are deleted" % del_num)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from celery.decorators import periodic_task
from celery.schedules import crontab
from django.conf import settings

from pipeline.log.models import LogEntry

logger = logging.getLogger(__name__)


@periodic_task(run_every=(crontab(minute=0, hour=0)), ignore_result=True)
def clean_expired_log():
    expired_interval = getattr(settings, "LOG_PERSISTENT_DAYS", None)

    if expired_interval is None:
        expired_interval = 30
        logger.warning("LOG_PERSISTENT_DAYS are not found in settings, use default value: 30")

    del_num = LogEntry.objects.delete_expired_log(expired_interval)
    logger.info("%s log entry are deleted" % del_num)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-11-09 11:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline_external_plugins", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="s3source", name="source_dir", field=models.TextField(default="", verbose_name="源目录名"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import queue

from django.utils.translation import ugettext_lazy as _

from pipeline import exceptions
from pipeline.core.constants import PE
from pipeline.engine.utils import Stack
from pipeline.validators.utils import get_node_for_sequence, get_nodes_dict

STREAM = "stream"
P_STREAM = "p_stream"
P = "p"
MAIN_STREAM = "main"

PARALLEL_GATEWAYS = {PE.ParallelGateway, PE.ConditionalParallelGateway}


def not_in_parallel_gateway(gateway_stack, start_from=None):
    """
    check whether there is parallel gateway in stack from specific gateway
    :param gateway_stack:
    :param start_from:
    :return:
    """
    start = 0
    if start_from:
        id_stack = [g[PE.id] for g in gateway_stack]
        start = id_stack.index(start_from)

    for i in range(start, len(gateway_stack)):
        gateway = gateway_stack[i]
        if gateway[PE.type] in PARALLEL_GATEWAYS:
            return False
    return True


def matched_in_prev_blocks(gid, current_start, block_nodes):
    """
    check whether gateway with gid is matched in previous block
    :param gid:
    :param current_start:
    :param block_nodes:
    :return:
    """
    prev_nodes = set()
    for prev_start, nodes in list(block_nodes.items()):
        if prev_start == current_start:
            continue
        prev_nodes.update(nodes)

    return gid in prev_nodes


def can_shared(current_gateway_id, converge_id, converged, gateways) -> bool:
    """
    check whether a converge gateway can be shared by current_gateway
    :param converge_id: converge gateway id
    :param converged: converge gateway converged gateway map
                      e.g. {"converge_gateway_id": ["parallel_gateway1", "parallel_gateway2]}
    :param gateways: gateway map
    """
    # exclusive gateway can share another block converge gateway
    if gateways[current_gateway_id][PE.type] not in PARALLEL_GATEWAYS:
        return True

    # check whether another parallel gateway in converged
    for gateway_id in converged.get(converge_id, []):
        if gateway_id == current_gateway_id:
            continue
        if gateways[gateway_id][PE.type] in PARALLEL_GATEWAYS:
            return False
    return True


def match_converge(
    converges,
    gateways,
    cur_index,
    end_event_id,
    block_start,
    block_nodes,
    converged,
    dist_from_start,
    converge_in_len,
    stack=None,
):
    """
    find converge for parallel and exclusive in blocks, and check sanity of gateway
    :param converges:
    :param gateways:
    :param cur_index:
    :param end_event_id:
    :param block_start:
    :param block_nodes:
    :param converged:
    :param dist_from_start:
    :param stack:
    :param converge_in_len:
    :return:
    """

    if stack is None:
        stack = Stack()

    if cur_index not in gateways:
        return None, False

    # return if this node is already matched
    if gateways[cur_index]["match"]:
        return gateways[cur_index]["match"], gateways[cur_index]["share_converge"]

    current_gateway = gateways[cur_index]
    target = gateways[cur_index][PE.target]
    stack.push(gateways[cur_index])
    stack_id_set = {g[PE.id] for g in stack}

    # find closest converge recursively
    for i in range(len(target)):

        # do not process prev blocks nodes
        if matched_in_prev_blocks(target[i], block_start, block_nodes):
            target[i] = None
            continue

        block_nodes[block_start].add(target[i])

        # do not find self's converge node again
        while target[i] in gateways and target[i] != current_gateway[PE.id]:

            if target[i] in stack_id_set:
                # return to previous gateway

                if not_in_parallel_gateway(stack, start_from=target[i]):
                    # do not trace back
                    target[i] = None
                    break
                else:
                    raise exceptions.ConvergeMatchError(cur_index, _("并行网关中的分支网关必须将所有分支汇聚到一个汇聚网关"))

            converge_id, shared = match_converge(
                converges=converges,
                gateways=gateways,
                cur_index=target[i],
                end_event_id=end_event_id,
                block_start=block_start,
                block_nodes=block_nodes,
                stack=stack,
                converged=converged,
                dist_from_start=dist_from_start,
                converge_in_len=converge_in_len,
            )
            if converge_id:
                target[i] = converge_id

                if not shared or (
                    shared
                    and not can_shared(
                        current_gateway_id=cur_index, converge_id=converge_id, converged=converged, gateways=gateways
                    )
                ):
                    # try to get next node fo converge which is not shared
                    target[i] = converges[converge_id][PE.target][0]

            else:
                # can't find corresponding converge gateway, which means this gateway will reach end event directly
                target[i] = end_event_id

        if target[i] in converges and dist_from_start[target[i]] < dist_from_start[cur_index]:
            # do not match previous converge
            target[i] = None

    stack.pop()

    is_exg = current_gateway[PE.type] == PE.ExclusiveGateway
    converge_id = None
    shared = False
    cur_to_converge = len(target)
    converge_end = False

    # gateway match validation
    for i in range(len(target)):

        # mark first converge
        if target[i] in converges and not converge_id:
            converge_id = target[i]

        # same converge node
        elif target[i] in converges and converge_id == target[i]:
            pass

        # exclusive gateway point to end
        elif is_exg and target[i] == end_event_id:
            if not_in_parallel_gateway(stack):
                converge_end = True
            else:
                raise exceptions.ConvergeMatchError(cur_index, _("并行网关中的分支网关必须将所有分支汇聚到一个汇聚网关"))

        # exclusive gateway point back to self
        elif is_exg and target[i] == current_gateway[PE.id]:
            # not converge behavior
            cur_to_converge -= 1
            pass

        # exclusive gateway converge at different converge gateway
        elif is_exg and target[i] in converges and converge_id != target[i]:
            raise exceptions.ConvergeMatchError(cur_index, _("分支网关的所有分支第一个遇到的汇聚网关必须是同一个"))

        # meet previous node
        elif is_exg and target[i] is None:
            # not converge behavior
            cur_to_converge -= 1
            pass

        # invalid cases
        else:
            raise exceptions.ConvergeMatchError(
                cur_index, _("非法网关, 请检查其分支是否符合规则, cur_index: %s, targets: %s" % (cur_index, target))
            )

    if is_exg:
        if converge_id in converges:
            # this converge is shared by multiple gateway
            # only compare to the number of positive incoming
            shared = converge_in_len[converge_id] > cur_to_converge or converge_id in converged
    else:
        # for parallel gateway

        converge_incoming = len(converges[converge_id][PE.incoming])
        gateway_outgoing = len(target)

        if converge_incoming > gateway_outgoing:
            for gateway_id in converged.get(converge_id, []):
                # find another parallel gateway
                if gateways[gateway_id][PE.type] in PARALLEL_GATEWAYS:
                    raise exceptions.ConvergeMatchError(converge_id, _("汇聚网关只能汇聚来自同一个并行网关的分支"))

            shared = True

        elif converge_incoming < gateway_outgoing:
            raise exceptions.ConvergeMatchError(converge_id, _("汇聚网关没有汇聚其对应的并行网关的所有分支"))

    current_gateway["match"] = converge_id
    current_gateway["share_converge"] = shared
    current_gateway["converge_end"] = converge_end

    converged.setdefault(converge_id, []).append(current_gateway[PE.id])
    block_nodes[block_start].add(current_gateway[PE.id])

    return converge_id, shared


def distance_from(origin, node, tree, marked, visited=None):
    """
    get max distance from origin to node
    :param origin:
    :param node:
    :param tree:
    :param marked:
    :param visited:
    :return:
    """
    if visited is None:
        visited = set()

    if node[PE.id] in marked:
        return marked[node[PE.id]]

    if node[PE.id] == origin[PE.id]:
        return 0

    if node[PE.id] in visited:
        # do not trace circle
        return None

    visited.add(node[PE.id])

    incoming_dist = []
    for incoming in node[PE.incoming]:
        prev_node = get_node_for_sequence(incoming, tree, PE.source)

        # get incoming node's distance recursively
        dist = distance_from(origin=origin, node=prev_node, tree=tree, marked=marked, visited=visited)

        # if this incoming do not trace back to current node
        if dist is not None:
            incoming_dist.append(dist + 1)

    if not incoming_dist:
        return None

    # get max distance
    res = max(incoming_dist)
    marked[node[PE.id]] = res
    return res


def validate_gateways(tree):
    """
    check sanity of gateways and find their converge gateway
    :param tree:
    :return:
    """
    converges = {}
    gateways = {}
    all = {}
    distances = {}
    converge_positive_in = {}
    process_order = []

    # data preparation
    for i, item in list(tree[PE.gateways].items()):
        node = {
            PE.incoming: item[PE.incoming] if isinstance(item[PE.incoming], list) else [item[PE.incoming]],
            PE.outgoing: item[PE.outgoing] if isinstance(item[PE.outgoing], list) else [item[PE.outgoing]],
            PE.type: item[PE.type],
            PE.target: [],
            PE.source: [],
            PE.id: item[PE.id],
            "match": None,
        }

        # find all first reach nodes(ConvergeGateway, ExclusiveGateway, ParallelGateway, EndEvent)
        # which is not ServiceActivity for each gateway
        for index in node[PE.outgoing]:
            index = tree[PE.flows][index][PE.target]
            while index in tree[PE.activities]:
                index = tree[PE.flows][tree[PE.activities][index][PE.outgoing]][PE.target]

            # append this node's id to current gateway's target list
            node[PE.target].append(index)

        # get current node's distance from start event
        if not distance_from(node=node, origin=tree[PE.start_event], tree=tree, marked=distances):
            raise exceptions.ConvergeMatchError(node[PE.id], _("无法获取该网关距离开始节点的距离"))

        if item[PE.type] == PE.ConvergeGateway:
            converges[i] = node
        else:
            process_order.append(i)
            gateways[i] = node

        all[i] = node

    # calculate positive incoming number for converge
    for nid, node in list(all.items()):
        for t in node[PE.target]:
            if t in converges and distances[t] > distances[nid]:
                converge_positive_in[t] = converge_positive_in.setdefault(t, 0) + 1

    process_order.sort(key=lambda gid: distances[gid])
    end_event_id = tree[PE.end_event][PE.id]
    converged = {}
    block_nodes = {}
    visited = set()

    # process in distance order
    for gw in process_order:
        if gw in visited or "match" in gw:
            continue
        visited.add(gw)

        block_nodes[gw] = set()

        match_converge(
            converges=converges,
            gateways=gateways,
            cur_index=gw,
            end_event_id=end_event_id,
            converged=converged,
            block_start=gw,
            block_nodes=block_nodes,
            dist_from_start=distances,
            converge_in_len=converge_positive_in,
        )

    # set converge gateway
    for i in gateways:
        if gateways[i]["match"]:
            tree[PE.gateways][i][PE.converge_gateway_id] = gateways[i]["match"]

    return converged


def blend(source, target, custom_stream=None):
    """
    blend source and target streams
    :param source:
    :param target:
    :param custom_stream:
    :return:
    """

    if custom_stream:
        # use custom stream instead of source's stream
        if isinstance(custom_stream, set):
            for stream in custom_stream:
                target[STREAM].add(stream)
        else:
            target[STREAM].add(custom_stream)

        return

    if len(source[STREAM]) == 0:
        raise exceptions.InvalidOperationException("stream validation error, node(%s) stream is empty" % source[PE.id])

    # blend
    for s in source[STREAM]:
        target[STREAM].add(s)


def streams_for_parallel(p):
    streams = set()
    for i, target_id in enumerate(p[PE.target]):
        streams.add("{}_{}".format(p[PE.id], i))

    return streams


def flowing(where, to, parallel_converges):
    """
    mark target's stream from target
    :param where:
    :param to:
    :param parallel_converges:
    :return:
    """
    is_parallel = where[PE.type] in PARALLEL_GATEWAYS

    stream = None
    if is_parallel:
        # add parallel's stream to its converge
        parallel_converge = to[where[PE.converge_gateway_id]]
        blend(source=where, target=parallel_converge, custom_stream=stream)

        if len(parallel_converge[STREAM]) > 1:
            raise exceptions.StreamValidateError(node_id=parallel_converge)

    # flow to target
    for i, target_id in enumerate(where[PE.target]):
        target = to[target_id]
        fake = False

        # generate different stream
        if is_parallel:
            stream = "{}_{}".format(where[PE.id], i)

        if target_id in parallel_converges:

            is_valid_branch = where[STREAM].issubset(parallel_converges[target_id][P_STREAM])
            is_direct_connect = where.get(PE.converge_gateway_id) == target_id

            if is_valid_branch or is_direct_connect:
                # do not flow when branch of parallel converge to its converge gateway
                fake = True

        if not fake:
            blend(source=where, target=target, custom_stream=stream)

        # sanity check
        if len(target[STREAM]) != 1:
            raise exceptions.StreamValidateError(node_id=target_id)


def validate_stream(tree):
    """
    validate flow stream
    :param tree: pipeline tree
    :return:
    """
    # data preparation
    start_event_id = tree[PE.start_event][PE.id]
    end_event_id = tree[PE.end_event][PE.id]
    nodes = get_nodes_dict(tree)
    nodes[start_event_id][STREAM] = {MAIN_STREAM}
    nodes[end_event_id][STREAM] = {MAIN_STREAM}
    parallel_converges = {}
    visited = set({})

    for nid, node in list(nodes.items()):
        node.setdefault(STREAM, set())

        # set allow streams for parallel's converge
        if node[PE.type] in PARALLEL_GATEWAYS:
            parallel_converges[node[PE.converge_gateway_id]] = {P_STREAM: streams_for_parallel(node), P: nid}

    # build stream from start
    node_queue = queue.Queue()
    node_queue.put(nodes[start_event_id])
    while not node_queue.empty():

        # get node
        node = node_queue.get()

        if node[PE.id] in visited:
            # flow again to validate stream, but do not add target to queue
            flowing(where=node, to=nodes, parallel_converges=parallel_converges)
            continue

        # add to queue
        for target_id in node[PE.target]:
            node_queue.put(nodes[target_id])

        # mark as visited
        visited.add(node[PE.id])

        # flow
        flowing(where=node, to=nodes, parallel_converges=parallel_converges)

    # data clean
    for nid, n in list(nodes.items()):
        if len(n[STREAM]) != 1:
            raise exceptions.StreamValidateError(node_id=nid)

        # replace set to str
        n[STREAM] = n[STREAM].pop()

    # isolate node check
    for __, node in list(nodes.items()):
        if not node[STREAM]:
            raise exceptions.IsolateNodeError()

    return nodes



# Generated by Django 2.2.6 on 2020-11-02 13:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("periodic_task", "0005_migrate_task"),
    ]

    operations = [
        migrations.AlterField(
            model_name="periodictask",
            name="celery_task",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="django_celery_beat.PeriodicTask",
                verbose_name="celery 周期任务实例",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pipeline',
 'pipeline.builder',
 'pipeline.builder.flow',
 'pipeline.celery',
 'pipeline.component_framework',
 'pipeline.component_framework.components',
 'pipeline.component_framework.components.collections',
 'pipeline.component_framework.management',
 'pipeline.component_framework.management.commands',
 'pipeline.component_framework.migrations',
 'pipeline.components',
 'pipeline.components.collections',
 'pipeline.conf',
 'pipeline.contrib',
 'pipeline.contrib.external_plugins',
 'pipeline.contrib.external_plugins.migrations',
 'pipeline.contrib.external_plugins.models',
 'pipeline.contrib.external_plugins.utils',
 'pipeline.contrib.external_plugins.utils.importer',
 'pipeline.contrib.periodic_task',
 'pipeline.contrib.periodic_task.djcelery',
 'pipeline.contrib.periodic_task.migrations',
 'pipeline.contrib.periodic_task.signals',
 'pipeline.contrib.statistics',
 'pipeline.contrib.statistics.migrations',
 'pipeline.contrib.statistics.signals',
 'pipeline.core',
 'pipeline.core.data',
 'pipeline.core.flow',
 'pipeline.core.flow.activity',
 'pipeline.core.signals',
 'pipeline.django_signal_valve',
 'pipeline.django_signal_valve.migrations',
 'pipeline.engine',
 'pipeline.engine.conf',
 'pipeline.engine.core',
 'pipeline.engine.core.data',
 'pipeline.engine.core.handlers',
 'pipeline.engine.core.handlers.endevent',
 'pipeline.engine.health',
 'pipeline.engine.health.zombie',
 'pipeline.engine.migrations',
 'pipeline.engine.models',
 'pipeline.engine.signals',
 'pipeline.eri',
 'pipeline.eri.celery',
 'pipeline.eri.imp',
 'pipeline.eri.migrations',
 'pipeline.log',
 'pipeline.log.migrations',
 'pipeline.management',
 'pipeline.management.commands',
 'pipeline.migrations',
 'pipeline.parser',
 'pipeline.service',
 'pipeline.service.pipeline_engine_adapter',
 'pipeline.signals',
 'pipeline.templates',
 'pipeline.templates.create_plugins_app',
 'pipeline.utils',
 'pipeline.utils.boolrule',
 'pipeline.utils.mako_utils',
 'pipeline.validators',
 'pipeline.variable_framework',
 'pipeline.variable_framework.management',
 'pipeline.variable_framework.management.commands',
 'pipeline.variable_framework.migrations',
 'pipeline.variable_framework.signals']

package_data = \
{'': ['*']}

install_requires = \
['Django>=2.2,<4.0',
 'Mako>=1.1.4,<2.0.0',
 'Werkzeug>=1.0.0,<2.0.0',
 'bamboo-engine==2.2.1',
 'boto3>=1.9.130,<2.0.0',
 'celery>=4.4.0,<5.0.0',
 'django-celery-beat>=2.1.0,<3.0.0',
 'django-timezone-field>=4.0,<5.0',
 'jsonschema>=2.5.1,<3.0.0',
 'prometheus-client>=0.9.0,<0.10.0',
 'pyparsing>=2.2.0,<3.0.0',
 'pytz==2019.3',
 'redis-py-cluster==2.1.0',
 'redis>=3.2.0,<4.0.0',
 'requests>=2.22.0,<3.0.0',
 'ujson>=4.1.0,<4.2.0']

setup_kwargs = {
    'name': 'bamboo-pipeline',
    'version': '3.17.1',
    'description': 'runtime for bamboo-engine base on Django and Celery',
    'long_description': None,
    'author': 'homholueng',
    'author_email': 'homholueng@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.8',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from copy import deepcopy

from django.utils.module_loading import import_string

from pipeline import exceptions
from pipeline.component_framework.library import ComponentLibrary
from pipeline.core.constants import PE
from pipeline.core.data.base import DataObject
from pipeline.core.data.context import Context
from pipeline.core.data.converter import get_variable
from pipeline.core.data.hydration import hydrate_node_data, hydrate_subprocess_context
from pipeline.core.flow import (
    Condition,
    ConditionalParallelGateway,
    ConvergeGateway,
    ExclusiveGateway,
    FlowNodeClsFactory,
    ParallelGateway,
    SequenceFlow,
)
from pipeline.core.pipeline import Pipeline, PipelineSpec
from pipeline.validators.base import validate_pipeline_tree


def classify_inputs(pipeline_inputs, params, is_subprocess, root_pipeline_params=None):
    """
    @summary: classify pipeline inputs into different parts
    @param pipeline_inputs: pipeline or subprocess inputs
    @param params: pipeline or subprocess params, which can cover item whose is_param is True in inputs
    @param is_subprocess: whether pipeline is root or subprocess
    @param root_pipeline_params: root pipeline params which should deliver to all subprocess
    @return:
    """
    # data from activity outputs
    act_outputs = {}
    # params should deliver to son subprocess
    subprocess_params = {}
    # context scope to resolving inputs
    scope_info = deepcopy(root_pipeline_params)
    for key, info in list(pipeline_inputs.items()):
        source_act = info.get(PE.source_act)
        if isinstance(source_act, str):
            act_outputs.setdefault(info[PE.source_act], {}).update({info[PE.source_key]: key})
            continue
        elif isinstance(source_act, list):
            for source_info in source_act:
                act_outputs.setdefault(source_info[PE.source_act], {}).update({source_info[PE.source_key]: key})

        is_param = info.get(PE.is_param, False)
        info = params.get(key, info) if is_param else info
        if is_subprocess and is_param:
            subprocess_params.update({key: info})
            continue

        scope_info.update({key: info})
    result = {"act_outputs": act_outputs, "scope_info": scope_info, "subprocess_params": subprocess_params}
    return result


class PipelineParser(object):
    def __init__(self, pipeline_tree, cycle_tolerate=False):
        validate_pipeline_tree(pipeline_tree, cycle_tolerate=cycle_tolerate)
        self.pipeline_tree = deepcopy(pipeline_tree)
        self.cycle_tolerate = cycle_tolerate

    def parse(self, root_pipeline_data=None, root_pipeline_context=None):
        """
        @summary: parse pipeline json tree to object with root data
        @param root_pipeline_data: like business info or operator, which can be accessed by parent_data in
            Component.execute
        @param root_pipeline_context: params for pipeline to resolving inputs data
        @return:
        """
        return self._parse(root_pipeline_data, root_pipeline_context)

    def _parse(
        self, root_pipeline_data=None, root_pipeline_params=None, params=None, is_subprocess=False, parent_context=None
    ):
        """
        @summary: parse pipeline and subprocess recursively
        @param root_pipeline_data: root data from root pipeline parsing, witch will be passed to subprocess recursively
        @param root_pipeline_params: params from root pipeline for all subprocess
        @param params: params from parent for son subprocess
        @param is_subprocess: whither is subprocess
        @param parent_context: parent context for activity of subprocess to resolving inputs
        @return: Pipeline object
        """
        if root_pipeline_data is None:
            root_pipeline_data = {}
        if root_pipeline_params is None:
            root_pipeline_params = {}
        if params is None:
            params = {}

        pipeline_inputs = self.pipeline_tree[PE.data][PE.inputs]
        classification = classify_inputs(pipeline_inputs, params, is_subprocess, root_pipeline_params)

        output_keys = self.pipeline_tree[PE.data][PE.outputs]
        context = Context(classification["act_outputs"], output_keys)
        for key, info in list(classification["scope_info"].items()):
            var = get_variable(key, info, context, root_pipeline_data)
            context.set_global_var(key, var)

        pipeline_data = deepcopy(root_pipeline_data)
        if is_subprocess:
            if parent_context is None:
                raise exceptions.DataTypeErrorException("parent context of subprocess cannot be none")
            for key, info in list(classification["subprocess_params"].items()):
                var = get_variable(key, info, parent_context, pipeline_data)
                pipeline_data.update({key: var})

        start = self.pipeline_tree[PE.start_event]
        start_cls = FlowNodeClsFactory.get_node_cls(start[PE.type])
        if "pre_render_keys" in self.pipeline_tree[PE.data]:
            start_event = start_cls(
                id=start[PE.id],
                name=start[PE.name],
                data=DataObject({"pre_render_keys": self.pipeline_tree[PE.data][PE.pre_render_keys]}),
            )
        else:
            start_event = start_cls(id=start[PE.id], name=start[PE.name])

        end = self.pipeline_tree[PE.end_event]
        end_cls = FlowNodeClsFactory.get_node_cls(end[PE.type])
        end_event = end_cls(id=end[PE.id], name=end[PE.name], data=DataObject({}))

        acts = self.pipeline_tree[PE.activities]
        act_objs = []
        for act in list(acts.values()):
            act_cls = FlowNodeClsFactory.get_node_cls(act[PE.type])
            if act[PE.type] == PE.ServiceActivity:
                component = ComponentLibrary.get_component(
                    component_code=act[PE.component][PE.code],
                    data_dict=act[PE.component][PE.inputs],
                    version=act[PE.component].get(PE.version),
                )
                service = component.service()
                data = component.data_for_execution(context, pipeline_data)
                handler_path = act.get("failure_handler")
                failure_handler = import_string(handler_path) if handler_path else None
                act_objs.append(
                    act_cls(
                        id=act[PE.id],
                        service=service,
                        name=act[PE.name],
                        data=data,
                        error_ignorable=act.get(PE.error_ignorable, False),
                        skippable=act[PE.skippable] if PE.skippable in act else act.get(PE.skippable_old, True),
                        retryable=act[PE.retryable] if PE.retryable in act else act.get(PE.retryable_old, True),
                        timeout=act.get(PE.timeout),
                        failure_handler=failure_handler,
                    )
                )
            elif act[PE.type] == PE.SubProcess:
                sub_tree = act[PE.pipeline]
                params = act[PE.params]
                sub_parser = PipelineParser(pipeline_tree=sub_tree, cycle_tolerate=self.cycle_tolerate)
                act_objs.append(
                    act_cls(
                        id=act[PE.id],
                        pipeline=sub_parser._parse(
                            root_pipeline_data=root_pipeline_data,
                            root_pipeline_params=root_pipeline_params,
                            params=params,
                            is_subprocess=True,
                            parent_context=context,
                        ),
                        name=act[PE.name],
                    )
                )
            else:
                raise exceptions.FlowTypeError("Unknown Activity type: %s" % act[PE.type])

        gateways = self.pipeline_tree[PE.gateways]
        flows = self.pipeline_tree[PE.flows]
        gateway_objs = []
        for gw in list(gateways.values()):
            gw_cls = FlowNodeClsFactory.get_node_cls(gw[PE.type])
            if gw[PE.type] in {PE.ParallelGateway, PE.ConditionalParallelGateway}:
                gateway_objs.append(
                    gw_cls(id=gw[PE.id], converge_gateway_id=gw[PE.converge_gateway_id], name=gw[PE.name])
                )
            elif gw[PE.type] in {PE.ExclusiveGateway, PE.ConvergeGateway}:
                gateway_objs.append(gw_cls(id=gw[PE.id], name=gw[PE.name]))
            else:
                raise exceptions.FlowTypeError("Unknown Gateway type: %s" % gw[PE.type])

        flow_objs_dict = {}
        for fl in list(flows.values()):
            flow_nodes = act_objs + gateway_objs
            if fl[PE.source] == start[PE.id]:
                source = start_event
            else:
                source = [x for x in flow_nodes if x.id == fl[PE.source]][0]
            if fl[PE.target] == end[PE.id]:
                target = end_event
            else:
                target = [x for x in flow_nodes if x.id == fl[PE.target]][0]
            flow_objs_dict[fl[PE.id]] = SequenceFlow(fl[PE.id], source, target)
        flow_objs = list(flow_objs_dict.values())

        # add incoming and outgoing flow to acts
        if not isinstance(start[PE.outgoing], list):
            start[PE.outgoing] = [start[PE.outgoing]]
        for outgoing_id in start[PE.outgoing]:
            start_event.outgoing.add_flow(flow_objs_dict[outgoing_id])

        if not isinstance(end[PE.incoming], list):
            end[PE.incoming] = [end[PE.incoming]]
        for incoming_id in end[PE.incoming]:
            end_event.incoming.add_flow(flow_objs_dict[incoming_id])

        for act in act_objs:
            incoming = acts[act.id][PE.incoming]
            if isinstance(incoming, list):
                for s in incoming:
                    act.incoming.add_flow(flow_objs_dict[s])
            else:
                act.incoming.add_flow(flow_objs_dict[incoming])

            act.outgoing.add_flow(flow_objs_dict[acts[act.id][PE.outgoing]])

        for gw in gateway_objs:
            if isinstance(gw, ExclusiveGateway) or isinstance(gw, ConditionalParallelGateway):
                for flow_id, con in list(gateways[gw.id][PE.conditions].items()):
                    con_obj = Condition(con[PE.evaluate], flow_objs_dict[flow_id])
                    gw.add_condition(con_obj)

                if isinstance(gateways[gw.id][PE.incoming], list):
                    for incoming_id in gateways[gw.id][PE.incoming]:
                        gw.incoming.add_flow(flow_objs_dict[incoming_id])
                else:
                    gw.incoming.add_flow(flow_objs_dict[gateways[gw.id][PE.incoming]])

                for outgoing_id in gateways[gw.id][PE.outgoing]:
                    gw.outgoing.add_flow(flow_objs_dict[outgoing_id])

            elif isinstance(gw, ParallelGateway):
                if isinstance(gateways[gw.id][PE.incoming], list):
                    for incoming_id in gateways[gw.id][PE.incoming]:
                        gw.incoming.add_flow(flow_objs_dict[incoming_id])
                else:
                    gw.incoming.add_flow(flow_objs_dict[gateways[gw.id][PE.incoming]])

                for outgoing_id in gateways[gw.id][PE.outgoing]:
                    gw.outgoing.add_flow(flow_objs_dict[outgoing_id])

            elif isinstance(gw, ConvergeGateway):
                for incoming_id in gateways[gw.id][PE.incoming]:
                    gw.incoming.add_flow(flow_objs_dict[incoming_id])
                gw.outgoing.add_flow(flow_objs_dict[gateways[gw.id][PE.outgoing]])

            else:
                raise exceptions.FlowTypeError("Unknown Gateway type: %s" % type(gw))

        context.duplicate_variables()
        pipeline_data = DataObject(pipeline_data)
        pipeline_spec = PipelineSpec(start_event, end_event, flow_objs, act_objs, gateway_objs, pipeline_data, context)
        return Pipeline(self.pipeline_tree[PE.id], pipeline_spec)

    def get_act(self, act_id, subprocess_stack=None, root_pipeline_data=None, root_pipeline_context=None):
        if subprocess_stack is None:
            subprocess_stack = []
        pipeline = self.parse(root_pipeline_data, root_pipeline_context)
        for sub_id in subprocess_stack:
            subprocess_act = [x for x in pipeline.spec.activities if x.id == sub_id][0]
            hydrate_subprocess_context(subprocess_act)
            pipeline = subprocess_act.pipeline
        act = [x for x in pipeline.spec.activities if x.id == act_id][0]
        return act

    def get_act_inputs(self, act_id, subprocess_stack=None, root_pipeline_data=None, root_pipeline_context=None):
        act = self.get_act(act_id, subprocess_stack, root_pipeline_data, root_pipeline_context)
        hydrate_node_data(act)
        inputs = act.data.inputs
        return inputs
